import warnings

import xarray as xr

from .np_deterministic import (
    _effective_sample_size,
    _linslope,
    _mae,
    _mape,
    _me,
    _median_absolute_error,
    _mse,
    _pearson_r,
    _pearson_r_eff_p_value,
    _pearson_r_p_value,
    _r2,
    _rmse,
    _smape,
    _spearman_r,
    _spearman_r_eff_p_value,
    _spearman_r_p_value,
)
from .utils import (
    _fail_if_dim_empty,
    _preprocess_dims,
    _preprocess_weights,
    _stack_input_if_needed,
)

__all__ = [
    "effective_sample_size",
    "linslope",
    "mae",
    "mape",
    "me",
    "median_absolute_error",
    "mse",
    "pearson_r",
    "pearson_r_eff_p_value",
    "pearson_r_p_value",
    "r2",
    "rmse",
    "smape",
    "spearman_r",
    "spearman_r_eff_p_value",
    "spearman_r_p_value",
]


def _determine_input_core_dims(dim, weights):
    """
    Determine input_core_dims based on type of dim and weights.

    Parameters
    ----------
    dim : str, list
        The dimension(s) to apply the metric along.
    weights : xarray.Dataset or xarray.DataArray or None
        Weights matching dimensions of ``dim`` to apply during the function.

    Returns
    -------
    list of lists
        input_core_dims used for xr.apply_ufunc.
    """
    if not isinstance(dim, list):
        dim = [dim]
    # build input_core_dims depending on weights
    if weights is None:
        input_core_dims = [dim, dim, []]
    else:
        input_core_dims = [dim, dim, dim]
    return input_core_dims


def linslope(a, b, dim=None, weights=None, skipna=False, keep_attrs=False):
    """Slope of linear fit.

    .. math::
        s_{ab} = \\frac{ \\sum_{i=i}^{n} (a_{i} - \\bar{a}) (b_{i} - \\bar{b}) }
                 { \\sum_{i=1}^{n} (a_{i} - \\bar{a})^{2} }

    Parameters
    ----------
    a : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    b : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    dim : str, list
        The dimension(s) to apply the slope of linear fit along. Note that this
        dimension will be reduced as a result. Defaults to None reducing all
        dimensions.
    weights : xarray.Dataset or xarray.DataArray or None
        Weights matching dimensions of ``dim`` to apply during the function.
    skipna : bool
        If True, skip NaNs when computing function.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied from the first input to the
        new one. If False (default), the new object will be returned without
        attributes.

    Returns
    -------
    xarray.DataArray or xarray.Dataset
        Slope of linear fit.

    See Also
    --------
    scipy.stats.linregress

    Examples
    --------
    >>> a = xr.DataArray(np.random.rand(5, 3, 3),
    ...                  dims=['time', 'x', 'y'])
    >>> b = xr.DataArray(np.random.rand(5, 3, 3),
    ...                  dims=['time', 'x', 'y'])
    >>> xs.linslope(a, b, dim='time')
    <xarray.DataArray (x: 3, y: 3)>
    array([[-0.30948771, -0.21562529, -0.63141304],
           [ 0.31446077,  2.23858011,  0.44743617],
           [-0.22243944,  0.47034784,  1.08512859]])
    Dimensions without coordinates: x, y
    """
    _fail_if_dim_empty(dim)
    dim, _ = _preprocess_dims(dim, a)
    a, b = xr.broadcast(a, b, exclude=dim)
    a, b, new_dim, weights = _stack_input_if_needed(a, b, dim, weights)
    weights = _preprocess_weights(a, dim, new_dim, weights)

    input_core_dims = _determine_input_core_dims(new_dim, weights)

    return xr.apply_ufunc(
        _linslope,
        a,
        b,
        weights,
        input_core_dims=input_core_dims,
        kwargs={"axis": -1, "skipna": skipna},
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=keep_attrs,
    )


def pearson_r(a, b, dim=None, weights=None, skipna=False, keep_attrs=False):
    """Pearson's correlation coefficient.

    .. math::
        r_{ab} = \\frac{ \\sum_{i=i}^{n} (a_{i} - \\bar{a}) (b_{i} - \\bar{b}) }
                 {\\sqrt{ \\sum_{i=1}^{n} (a_{i} - \\bar{a})^{2} }
                  \\sqrt{ \\sum_{i=1}^{n} (b_{i} - \\bar{b})^{2} }}

    Parameters
    ----------
    a : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    b : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    dim : str, list
        The dimension(s) to apply the correlation along. Note that this dimension will
        be reduced as a result. Defaults to None reducing all dimensions.
    weights : xarray.Dataset or xarray.DataArray or None
        Weights matching dimensions of ``dim`` to apply during the function.
    skipna : bool
        If True, skip NaNs when computing function.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied
        from the first input to the new one.
        If False (default), the new object will
        be returned without attributes.

    Returns
    -------
    xarray.DataArray or xarray.Dataset
        Pearson's correlation coefficient.

    See Also
    --------
    scipy.stats.pearsonr

    References
    ----------
    https://en.wikipedia.org/wiki/Pearson_correlation_coefficient

    Examples
    --------
    >>> a = xr.DataArray(np.random.rand(5, 3, 3),
    ...                  dims=['time', 'x', 'y'])
    >>> b = xr.DataArray(np.random.rand(5, 3, 3),
    ...                  dims=['time', 'x', 'y'])
    >>> xs.pearson_r(a, b, dim='time')
    <xarray.DataArray (x: 3, y: 3)>
    array([[-0.17455755, -0.26648379, -0.74265833],
           [ 0.32535918,  0.42496646,  0.1940647 ],
           [-0.3203094 ,  0.33207755,  0.89250429]])
    Dimensions without coordinates: x, y
    """
    _fail_if_dim_empty(dim)
    dim, _ = _preprocess_dims(dim, a)
    a, b = xr.broadcast(a, b, exclude=dim)
    a, b, new_dim, weights = _stack_input_if_needed(a, b, dim, weights)
    weights = _preprocess_weights(a, dim, new_dim, weights)

    input_core_dims = _determine_input_core_dims(new_dim, weights)

    return xr.apply_ufunc(
        _pearson_r,
        a,
        b,
        weights,
        input_core_dims=input_core_dims,
        kwargs={"axis": -1, "skipna": skipna},
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=keep_attrs,
    )


def pearson_r_p_value(a, b, dim=None, weights=None, skipna=False, keep_attrs=False):
    """2-tailed p-value associated with pearson's correlation coefficient.

    Parameters
    ----------
    a : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    b : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    dim : str, list
        The dimension(s) to apply the correlation along. Note that this dimension will
        be reduced as a result. Defaults to None reducing all dimensions.
    weights : xarray.Dataset or xarray.DataArray or None
        Weights matching dimensions of ``dim`` to apply during the function.
    skipna : bool
        If True, skip NaNs when computing function.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied
        from the first input to the new one.
        If False (default), the new object will
        be returned without attributes.

    Returns
    -------
    xarray.Dataset or xarray.DataArray
        2-tailed p-value of Pearson's correlation coefficient.

    See Also
    --------
    scipy.stats.pearsonr

    Examples
    --------
    >>> a = xr.DataArray(np.random.rand(5, 3, 3),
    ...                  dims=['time', 'x', 'y'])
    >>> b = xr.DataArray(np.random.rand(5, 3, 3),
    ...                  dims=['time', 'x', 'y'])
    >>> xs.pearson_r_p_value(a, b, dim='time')
    <xarray.DataArray (x: 3, y: 3)>
    array([[0.77888033, 0.66476199, 0.15051516],
           [0.59316935, 0.47567465, 0.75446898],
           [0.59925464, 0.58509064, 0.04161894]])
    Dimensions without coordinates: x, y
    """
    _fail_if_dim_empty(dim)
    dim, _ = _preprocess_dims(dim, a)
    a, b = xr.broadcast(a, b, exclude=dim)
    a, b, new_dim, weights = _stack_input_if_needed(a, b, dim, weights)
    weights = _preprocess_weights(a, dim, new_dim, weights)
    input_core_dims = _determine_input_core_dims(new_dim, weights)

    return xr.apply_ufunc(
        _pearson_r_p_value,
        a,
        b,
        weights,
        input_core_dims=input_core_dims,
        kwargs={"axis": -1, "skipna": skipna},
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=keep_attrs,
    )


def effective_sample_size(a, b, dim="time", skipna=False, keep_attrs=False):
    """Effective sample size for temporally correlated data.

    .. note::
        This metric should only be applied over the time dimension,
        since it is designed for temporal autocorrelation. Weights
        are not included due to the reliance on temporal
        autocorrelation.

    The effective sample size extracts the number of independent samples
    between two time series being correlated. This is derived by assessing
    the magnitude of the lag-1 autocorrelation coefficient in each of the time series
    being correlated. A higher autocorrelation induces a lower effective sample
    size which raises the correlation coefficient for a given p value.

     .. math::
        N_{eff} = N\\left( \\frac{1 -
                   \\rho_{f}\\rho_{o}}{1 + \\rho_{f}\\rho_{o}} \\right),

    where :math:`\\rho_{f}` and :math:`\\rho_{o}` are the lag-1 autocorrelation
    coefficients for the forecast and observations.

    Parameters
    ----------
    a : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    b : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    dim : str, list
        The dimension(s) to apply the function along. Note that this dimension will
        be reduced as a result. Defaults to None reducing all dimensions.
    skipna : bool
        If True, skip NaNs when computing function.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied
        from the first input to the new one.
        If False (default), the new object will
        be returned without attributes.

    Returns
    -------
    xarray.Dataset or xarray.DataArray
        Effective sample size.

    References
    ----------
    * Bretherton, Christopher S., et al. "The effective number of spatial degrees of
      freedom of a time-varying field." Journal of climate 12.7 (1999): 1990-2009.
    * Wilks, Daniel S. Statistical methods in the atmospheric sciences. Vol. 100.
      Academic press, 2011.

    Examples
    --------
    >>> a = xr.DataArray(np.random.rand(5, 3, 3),
    ...                  dims=['time', 'x', 'y'])
    >>> b = xr.DataArray(np.random.rand(5, 3, 3),
    ...                  dims=['time', 'x', 'y'])
    >>> xs.effective_sample_size(a, b, dim='time')
    <xarray.DataArray (x: 3, y: 3)>
    array([[4., 0., 4.],
           [3., 4., 4.],
           [3., 4., 2.]])
    Dimensions without coordinates: x, y
    """
    _fail_if_dim_empty(dim)
    dim, _ = _preprocess_dims(dim, a)
    a, b = xr.broadcast(a, b, exclude=dim)

    if len(dim) > 1:
        raise ValueError(
            "Effective sample size should only be applied to a singular time dimension."
        )
    else:
        new_dim = dim[0]
    if new_dim != "time":
        warnings.warn(
            f"{dim} is not 'time'. Make sure that you are applying this over a "
            f"temporal dimension."
        )

    return xr.apply_ufunc(
        _effective_sample_size,
        a,
        b,
        input_core_dims=[[new_dim], [new_dim]],
        kwargs={"axis": -1, "skipna": skipna},
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=keep_attrs,
    )


def pearson_r_eff_p_value(a, b, dim=None, skipna=False, keep_attrs=False):
    """
    2-tailed p-value associated with Pearson's correlation coefficient,
    accounting for autocorrelation.

    .. note::
        This metric should only be applied over the time dimension,
        since it is designed for temporal autocorrelation. Weights
        are not included due to the reliance on temporal
        autocorrelation.

    The effective p value is computed by replacing the sample size :math:`N` in the
    t-statistic with the effective sample size, :math:`N_{eff}`. The same Pearson
    product-moment correlation coefficient :math:`r` is used as when computing the
    standard p value.

    .. math::
        t = r\\sqrt{ \\frac{N_{eff} - 2}{1 - r^{2}} },

    where :math:`N_{eff}` is computed via the autocorrelation in the forecast and
    observations.

    .. math::
        N_{eff} = N\\left( \\frac{1 -
                   \\rho_{f}\\rho_{o}}{1 + \\rho_{f}\\rho_{o}} \\right),

    where :math:`\\rho_{f}` and :math:`\\rho_{o}` are the lag-1 autocorrelation
    coefficients for the forecast and observations.

    Parameters
    ----------
    a : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    b : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    dim : str, list
        The dimension(s) to compute the p value over. Note that this dimension will
        be reduced as a result. Defaults to None reducing all dimensions.
    skipna : bool
        If True, skip NaNs when computing function.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied
        from the first input to the new one.
        If False (default), the new object will
        be returned without attributes.

    Returns
    -------
    xarray.Dataset or xarray.DataArray
        2-tailed p-value of Pearson's correlation coefficient, accounting
        for autocorrelation.

    See Also
    --------
    scipy.stats.pearsonr

    References
    ----------
    * Bretherton, Christopher S., et al. "The effective number of spatial degrees of
      freedom of a time-varying field." Journal of climate 12.7 (1999): 1990-2009.
    * Wilks, Daniel S. Statistical methods in the atmospheric sciences. Vol. 100.
      Academic press, 2011.

    Examples
    --------
    >>> a = xr.DataArray(np.random.rand(5, 3, 3),
    ...                  dims=['time', 'x', 'y'])
    >>> b = xr.DataArray(np.random.rand(5, 3, 3),
    ...                  dims=['time', 'x', 'y'])
    >>> xs.pearson_r_eff_p_value(a, b, dim='time')
    <xarray.DataArray (x: 3, y: 3)>
    array([[0.82544245,        nan, 0.25734167],
           [0.78902959, 0.57503354, 0.8059353 ],
           [0.79242625, 0.66792245,        nan]])
    Dimensions without coordinates: x, y
    """
    _fail_if_dim_empty(dim)
    dim, _ = _preprocess_dims(dim, a)
    a, b = xr.broadcast(a, b, exclude=dim)

    if len(dim) > 1:
        raise ValueError(
            "Effective sample size should only be applied to a singular time dimension."
        )
    else:
        new_dim = dim[0]
    if new_dim != "time":
        warnings.warn(
            f"{dim} is not 'time'. Make sure that you are applying this over a "
            f"temporal dimension."
        )

    return xr.apply_ufunc(
        _pearson_r_eff_p_value,
        a,
        b,
        input_core_dims=[[new_dim], [new_dim]],
        kwargs={"axis": -1, "skipna": skipna},
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=keep_attrs,
    )


def spearman_r(a, b, dim=None, weights=None, skipna=False, keep_attrs=False):
    """Spearman's correlation coefficient.

    Parameters
    ----------
    a : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    b : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    dim : str, list
        The dimension(s) to apply the correlation along. Note that this dimension will
        be reduced as a result. Defaults to None reducing all dimensions.
    weights : xarray.Dataset or xarray.DataArray or None
        Weights matching dimensions of ``dim`` to apply during the function.
    skipna : bool
        If True, skip NaNs when computing function.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied
        from the first input to the new one.
        If False (default), the new object will
        be returned without attributes.

    Returns
    -------
    xarray.DataArray or xarray.Dataset
        Spearman's correlation coefficient.

    See Also
    --------
    scipy.stats.spearman_r

    References
    ----------
    https://github.com/scipy/scipy/blob/v1.3.1/scipy/stats/stats.py#L3613-L3764
    https://en.wikipedia.org/wiki/Spearman%27s_rank_correlation_coefficient

    Examples
    --------
    >>> a = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> b = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> xs.spearman_r(a, b, dim='time')
    <xarray.DataArray (x: 3, y: 3)>
    array([[-0.6, -0.5, -0.7],
           [ 0.4,  0.3,  0.3],
           [-0.3, -0.1,  0.9]])
    Dimensions without coordinates: x, y
    """
    _fail_if_dim_empty(dim)
    dim, _ = _preprocess_dims(dim, a)
    a, b = xr.broadcast(a, b, exclude=dim)
    a, b, new_dim, weights = _stack_input_if_needed(a, b, dim, weights)
    weights = _preprocess_weights(a, dim, new_dim, weights)
    input_core_dims = _determine_input_core_dims(new_dim, weights)

    return xr.apply_ufunc(
        _spearman_r,
        a,
        b,
        weights,
        input_core_dims=input_core_dims,
        kwargs={"axis": -1, "skipna": skipna},
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=keep_attrs,
    )


def spearman_r_p_value(a, b, dim=None, weights=None, skipna=False, keep_attrs=False):
    """2-tailed p-value associated with Spearman's correlation coefficient.

    Parameters
    ----------
    a : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    b : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    dim : str, list
        The dimension(s) to apply the correlation along. Note that this dimension will
        be reduced as a result. Defaults to None reducing all dimensions.
    weights : xarray.Dataset or xarray.DataArray or None
        Weights matching dimensions of ``dim`` to apply during the function.
    skipna : bool
        If True, skip NaNs when computing function.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied
        from the first input to the new one.
        If False (default), the new object will
        be returned without attributes.

    Returns
    -------
    xarray.Dataset or xarray.DataArray
        2-tailed p-value of Spearman's correlation coefficient.

    See Also
    --------
    scipy.stats.spearman_r

    Examples
    --------
    >>> a = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> b = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> xs.spearman_r_p_value(a, b, dim='time')
    <xarray.DataArray (x: 3, y: 3)>
    array([[0.28475698, 0.39100222, 0.1881204 ],
           [0.50463158, 0.62383766, 0.62383766],
           [0.62383766, 0.87288857, 0.03738607]])
    Dimensions without coordinates: x, y
    """
    _fail_if_dim_empty(dim)
    dim, _ = _preprocess_dims(dim, a)
    a, b = xr.broadcast(a, b, exclude=dim)
    a, b, new_dim, weights = _stack_input_if_needed(a, b, dim, weights)
    weights = _preprocess_weights(a, dim, new_dim, weights)
    input_core_dims = _determine_input_core_dims(new_dim, weights)

    return xr.apply_ufunc(
        _spearman_r_p_value,
        a,
        b,
        weights,
        input_core_dims=input_core_dims,
        kwargs={"axis": -1, "skipna": skipna},
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=keep_attrs,
    )


def spearman_r_eff_p_value(a, b, dim=None, skipna=False, keep_attrs=False):
    """
    2-tailed p-value associated with Spearman rank correlation coefficient,
    accounting for autocorrelation.

    .. note::
        This metric should only be applied over the time dimension,
        since it is designed for temporal autocorrelation. Weights
        are not included due to the reliance on temporal
        autocorrelation.

    The effective p value is computed by replacing the sample size :math:`N` in the
    t-statistic with the effective sample size, :math:`N_{eff}`. The same Spearman's
    rank correlation coefficient :math:`r` is used as when computing the standard p
    value.

    .. math::
        t = r\\sqrt{ \\frac{N_{eff} - 2}{1 - r^{2}} },

    where :math:`N_{eff}` is computed via the autocorrelation in the forecast and
    observations.

    .. math::
        N_{eff} = N\\left( \\frac{1 -
                   \\rho_{f}\\rho_{o}}{1 + \\rho_{f}\\rho_{o}} \\right),

    where :math:`\\rho_{f}` and :math:`\\rho_{o}` are the lag-1 autocorrelation
    coefficients for the forecast and observations.

    Parameters
    ----------
    a : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    b : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    dim : str, list
        The dimension(s) to compute the p value over. Note that this dimension will
        be reduced as a result. Defaults to None reducing all dimensions.
    skipna : bool
        If True, skip NaNs when computing function.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied
        from the first input to the new one.
        If False (default), the new object will
        be returned without attributes.

    Returns
    -------
    xarray.Dataset or xarray.DataArray
        2-tailed p-value of Spearman's correlation coefficient, accounting for
        autocorrelation.

    See Also
    --------
    scipy.stats.spearman_r

    References
    ----------
    * Bretherton, Christopher S., et al. "The effective number of spatial degrees of
      freedom of a time-varying field." Journal of climate 12.7 (1999): 1990-2009.
    * Wilks, Daniel S. Statistical methods in the atmospheric sciences. Vol. 100.
      Academic press, 2011.

    Examples
    --------
    >>> a = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> b = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> xs.spearman_r_eff_p_value(a, b, dim='time')
    <xarray.DataArray (x: 3, y: 3)>
    array([[0.4       ,        nan, 0.3       ],
           [0.73802024, 0.7       , 0.7       ],
           [0.80602663, 0.9       ,        nan]])
    Dimensions without coordinates: x, y
    """
    _fail_if_dim_empty(dim)
    dim, _ = _preprocess_dims(dim, a)
    a, b = xr.broadcast(a, b, exclude=dim)

    if len(dim) > 1:
        raise ValueError(
            "Effective sample size should only be applied to a singular time dimension."
        )
    else:
        new_dim = dim[0]
    if new_dim != "time":
        warnings.warn(
            f"{dim} is not 'time'. Make sure that you are applying this over a "
            f"temporal dimension."
        )

    return xr.apply_ufunc(
        _spearman_r_eff_p_value,
        a,
        b,
        input_core_dims=[[new_dim], [new_dim]],
        kwargs={"axis": -1, "skipna": skipna},
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=keep_attrs,
    )


def r2(a, b, dim=None, weights=None, skipna=False, keep_attrs=False):
    """R^2 (coefficient of determination) score.

    We first take the total sum of squares of our known vector, a.

    .. math::
        SS_{\\mathrm{tot}} = \\sum_{i=1}^{n} (a_{i} - \\bar{a})^{2}

    Next, we take the sum of squares of the error between our known vector
    a and the predicted vector, b.

    .. math::
        SS_{\\mathrm{res}} = \\sum_{i=1}^{n} (a_{i} - b_{i})^{2}

    Lastly we compute the coefficient of determiniation using these two
    terms.

    .. math::
        R^{2} = 1 - \\frac{SS_{\\mathrm{res}}}{SS_{\\mathrm{tot}}}

    .. note::
        The coefficient of determination is *not* symmetric. In other words,
        ``r2(a, b) != r2(b, a)``. Be careful and note that by our
        convention, ``b`` is the modeled/predicted vector and ``a`` is the
        observed vector.

    Parameters
    ----------
    a : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    b : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    dim : str, list
        The dimension(s) to apply the correlation along. Note that this dimension will
        be reduced as a result. Defaults to None reducing all dimensions.
    weights : xarray.Dataset or xarray.DataArray or None
        Weights matching dimensions of ``dim`` to apply during the function.
    skipna : bool
        If True, skip NaNs when computing function.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied
        from the first input to the new one.
        If False (default), the new object will
        be returned without attributes.

    Returns
    -------
    xarray.DataArray or xarray.Dataset
        R^2 (coefficient of determination) score.

    See Also
    --------
    sklearn.metrics.r2_score

    References
    ----------
    https://en.wikipedia.org/wiki/Coefficient_of_determination

    Examples
    --------
    >>> a = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> b = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> r2(a, b, dim='time')
    <xarray.DataArray (x: 3, y: 3)>
    array([[ -3.77828319,  -1.25687543,  -2.52495914],
           [ -0.67280201, -39.45271514,  -5.78241791],
           [ -1.66615797,  -1.56749317,   0.09843265]])
    Dimensions without coordinates: x, y
    """
    _fail_if_dim_empty(dim)
    dim, _ = _preprocess_dims(dim, a)
    a, b = xr.broadcast(a, b, exclude=dim)
    a, b, new_dim, weights = _stack_input_if_needed(a, b, dim, weights)
    weights = _preprocess_weights(a, dim, new_dim, weights)

    input_core_dims = _determine_input_core_dims(new_dim, weights)

    return xr.apply_ufunc(
        _r2,
        a,
        b,
        weights,
        input_core_dims=input_core_dims,
        kwargs={"axis": -1, "skipna": skipna},
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=keep_attrs,
    )


def me(a, b, dim=None, weights=None, skipna=False, keep_attrs=False):
    """Mean Error.

    .. math::
        \\mathrm{ME} = \\frac{1}{n}\\sum_{i=1}^{n}(a_{i} - b_{i})

    Parameters
    ----------
    a : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    b : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    dim : str, list
        The dimension(s) to apply the me along. Note that this dimension will
        be reduced as a result. Defaults to None reducing all dimensions.
    weights : xarray.Dataset or xarray.DataArray or None
        Weights matching dimensions of ``dim`` to apply during the function.
    skipna : bool
        If True, skip NaNs when computing function.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied
        from the first input to the new one.
        If False (default), the new object will
        be returned without attributes.

    Returns
    -------
    xarray.Dataset or xarray.DataArray
        Mean Error.

    Examples
    --------
    >>> a = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> b = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> me(a, b, dim='time')
    <xarray.DataArray (x: 3, y: 3)>
    array([[ 0.01748202, -0.14165293,  0.22455357],
           [ 0.13893709, -0.23513353, -0.18174132],
           [-0.29317762,  0.16887445, -0.17297527]])
    Dimensions without coordinates: x, y
    """
    dim, axis = _preprocess_dims(dim, a)
    a, b = xr.broadcast(a, b, exclude=dim)
    weights = _preprocess_weights(a, dim, dim, weights)
    input_core_dims = _determine_input_core_dims(dim, weights)

    return xr.apply_ufunc(
        _me,
        a,
        b,
        weights,
        input_core_dims=input_core_dims,
        kwargs={"axis": axis, "skipna": skipna},
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=keep_attrs,
    )


def rmse(a, b, dim=None, weights=None, skipna=False, keep_attrs=False):
    """Root Mean Squared Error.

    .. math::
        \\mathrm{RMSE} = \\sqrt{\\frac{1}{n}\\sum_{i=1}^{n}(a_{i} - b_{i})^{2}}

    Parameters
    ----------
    a : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    b : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    dim : str, list
        The dimension(s) to apply the rmse along. Note that this dimension will
        be reduced as a result. Defaults to None reducing all dimensions.
    weights : xarray.Dataset or xarray.DataArray or None
        Weights matching dimensions of ``dim`` to apply during the function.
    skipna : bool
        If True, skip NaNs when computing function.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied
        from the first input to the new one.
        If False (default), the new object will
        be returned without attributes.

    Returns
    -------
    xarray.Dataset or xarray.DataArray
        Root Mean Squared Error.

    See Also
    --------
    sklearn.metrics.mean_squared_error

    References
    ----------
    https://en.wikipedia.org/wiki/Root-mean-square_deviation

    Examples
    --------
    >>> a = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> b = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> xs.rmse(a, b, dim='time')
    <xarray.DataArray (x: 3, y: 3)>
    array([[0.30366872, 0.5147618 , 0.57410211],
           [0.2963848 , 0.37177283, 0.40563885],
           [0.55686111, 0.38189299, 0.21317579]])
    Dimensions without coordinates: x, y
    """
    dim, axis = _preprocess_dims(dim, a)
    a, b = xr.broadcast(a, b, exclude=dim)
    weights = _preprocess_weights(a, dim, dim, weights)
    input_core_dims = _determine_input_core_dims(dim, weights)

    return xr.apply_ufunc(
        _rmse,
        a,
        b,
        weights,
        input_core_dims=input_core_dims,
        kwargs={"axis": axis, "skipna": skipna},
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=keep_attrs,
    )


def mse(a, b, dim=None, weights=None, skipna=False, keep_attrs=False):
    """Mean Squared Error.

    .. math::
        \\mathrm{MSE} = \\frac{1}{n}\\sum_{i=1}^{n}(a_{i} - b_{i})^{2}

    Parameters
    ----------
    a : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    b : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    dim : str, list
        The dimension(s) to apply the mse along. Note that this dimension will
        be reduced as a result. Defaults to None reducing all dimensions.
    weights : xarray.Dataset or xarray.DataArray or None
        Weights matching dimensions of ``dim`` to apply during the function.
    skipna : bool
        If True, skip NaNs when computing function.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied
    --------
        from the first input to the new one.
        If False (default), the new object will
        be returned without attributes.

    Returns
    -------
    xarray.Dataset or xarray.DataArray
        Mean Squared Error.

    See Also
    --------
    sklearn.metrics.mean_squared_error

    References
    ----------
    https://en.wikipedia.org/wiki/Mean_squared_error

    Examples
    --------
    >>> a = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> b = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> xs.mse(a, b, dim='time')
    <xarray.DataArray (x: 3, y: 3)>
    array([[0.09221469, 0.26497971, 0.32959323],
           [0.08784395, 0.13821504, 0.16454288],
           [0.31009429, 0.14584225, 0.04544392]])
    Dimensions without coordinates: x, y
    """
    dim, axis = _preprocess_dims(dim, a)
    a, b = xr.broadcast(a, b, exclude=dim)
    weights = _preprocess_weights(a, dim, dim, weights)
    input_core_dims = _determine_input_core_dims(dim, weights)

    return xr.apply_ufunc(
        _mse,
        a,
        b,
        weights,
        input_core_dims=input_core_dims,
        kwargs={"axis": axis, "skipna": skipna},
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=keep_attrs,
    )


def mae(a, b, dim=None, weights=None, skipna=False, keep_attrs=False):
    """Mean Absolute Error.

    .. math::
        \\mathrm{MAE} = \\frac{1}{n}\\sum_{i=1}^{n}\\vert a - b\\vert

    Parameters
    ----------
    a : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    b : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    dim : str, list
        The dimension(s) to apply the mae along. Note that this dimension will
        be reduced as a result. Defaults to None reducing all dimensions.
    weights : xarray.Dataset or xarray.DataArray or None
        Weights matching dimensions of ``dim`` to apply during the function.
    skipna : bool
        If True, skip NaNs when computing function.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied
        from the first input to the new one.
        If False (default), the new object will
        be returned without attributes.

    Returns
    -------
    xarray.Dataset or xarray.DataArray
        Mean Absolute Error.

    See Also
    --------
    sklearn.metrics.mean_absolute_error

    References
    ----------
    https://en.wikipedia.org/wiki/Mean_absolute_error

    Examples
    --------
    >>> a = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> b = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> mae(a, b, dim='time')
    <xarray.DataArray (x: 3, y: 3)>
    array([[0.26014863, 0.40137207, 0.48871634],
           [0.18809417, 0.30197826, 0.2984658 ],
           [0.52934554, 0.19820357, 0.17335851]])
    Dimensions without coordinates: x, y
    """
    dim, axis = _preprocess_dims(dim, a)
    a, b = xr.broadcast(a, b, exclude=dim)
    weights = _preprocess_weights(a, dim, dim, weights)
    input_core_dims = _determine_input_core_dims(dim, weights)

    return xr.apply_ufunc(
        _mae,
        a,
        b,
        weights,
        input_core_dims=input_core_dims,
        kwargs={"axis": axis, "skipna": skipna},
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=keep_attrs,
    )


def median_absolute_error(a, b, dim=None, skipna=False, keep_attrs=False):
    """
    Median Absolute Error.

    .. math::
        \\mathrm{median}(\\vert a - b\\vert)

    Parameters
    ----------
    a : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    b : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    dim : str, list
        The dimension(s) to apply the median absolute error along.
        Note that this dimension will be reduced as a result.
        Defaults to None reducing all dimensions.
    skipna : bool
        If True, skip NaNs when computing function.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied
        from the first input to the new one.
        If False (default), the new object will
        be returned without attributes.

    Returns
    -------
    xarray.Dataset or xarray.DataArray
        Median Absolute Error.

    See Also
    --------
    sklearn.metrics.median_absolute_error

    Examples
    --------
    >>> a = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> b = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> xs.median_absolute_error(a, b, dim='time')
    <xarray.DataArray (x: 3, y: 3)>
    array([[0.28798217, 0.23322591, 0.62067468],
           [0.12146232, 0.20314509, 0.23442927],
           [0.59041981, 0.03289321, 0.21343494]])
    Dimensions without coordinates: x, y
    """
    dim, axis = _preprocess_dims(dim, a)
    a, b = xr.broadcast(a, b, exclude=dim)

    return xr.apply_ufunc(
        _median_absolute_error,
        a,
        b,
        input_core_dims=[dim, dim],
        kwargs={"axis": axis, "skipna": skipna},
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=keep_attrs,
    )


def mape(a, b, dim=None, weights=None, skipna=False, keep_attrs=False):
    """Mean Absolute Percentage Error.

    .. math::
        \\mathrm{MAPE} = \\frac{1}{n} \\sum_{i=1}^{n}
                         \\frac{\\vert a_{i} - b_{i} \\vert}
                               {max(\epsilon, \\vert a_{i} \\vert)}

    .. note::
        The percent error is calculated in reference to ``a``. Percent
        error is reported as decimal percent. I.e., a value of 1 is
        100%. :math:`\epsilon` is an arbitrary small yet strictly positive
        number to avoid undefined results when ``a`` is zero.

    Parameters
    ----------
    a : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
        (Truth which will be divided by)
    b : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    dim : str, list
        The dimension(s) to apply mape along. Note that this dimension will
        be reduced as a result. Defaults to None reducing all dimensions.
    weights : xarray.Dataset or xarray.DataArray or None
        Weights matching dimensions of ``dim`` to apply during the function.
    skipna : bool
        If True, skip NaNs when computing function.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied
        from the first input to the new one.
        If False (default), the new object will
        be returned without attributes.

    Returns
    -------
    xarray.Dataset or xarray.DataArray
        Mean Absolute Percentage Error.

    See Also
    --------
    sklearn.metrics.mean_absolute_percentage_error

    References
    ----------
    https://en.wikipedia.org/wiki/Mean_absolute_percentage_error

    Examples
    --------
    >>> a = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> b = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> xs.mape(a, b, dim='time')
    <xarray.DataArray (x: 3, y: 3)>
    array([[0.6268041 , 9.45134297, 3.28717608],
           [0.27099746, 1.58105176, 1.48258713],
           [6.55806162, 0.22271096, 0.39302745]])
    Dimensions without coordinates: x, y
    """
    dim, axis = _preprocess_dims(dim, a)
    a, b = xr.broadcast(a, b, exclude=dim)
    weights = _preprocess_weights(a, dim, dim, weights)
    input_core_dims = _determine_input_core_dims(dim, weights)

    return xr.apply_ufunc(
        _mape,
        a,
        b,
        weights,
        input_core_dims=input_core_dims,
        kwargs={"axis": axis, "skipna": skipna},
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=keep_attrs,
    )


def smape(a, b, dim=None, weights=None, skipna=False, keep_attrs=False):
    """Symmetric Mean Absolute Percentage Error.

    .. math::
        \\mathrm{SMAPE} = \\frac{1}{n} \\sum_{i=1}^{n}
                          \\frac{ \\vert a_{i} - b_{i} \\vert }
                          { \\vert a_{i} \\vert + \\vert b_{i} \\vert  }

    .. note::
        Percent error is reported as decimal percent. I.e., a value of 1 is
        100%.

    Parameters
    ----------
    a : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
        (Truth which will be divided by)
    b : xarray.Dataset or xarray.DataArray
        Labeled array(s) over which to apply the function.
    dim : str, list
        The dimension(s) to apply the smape along. Note that this dimension will
        be reduced as a result. Defaults to None reducing all dimensions.
    weights : xarray.Dataset or xarray.DataArray or None
        Weights matching dimensions of ``dim`` to apply during the function.
    skipna : bool
        If True, skip NaNs when computing function.
    keep_attrs : bool
        If True, the attributes (attrs) will be copied
        from the first input to the new one.
        If False (default), the new object will
        be returned without attributes.

    Returns
    -------
    xarray.Dataset or xarray.DataArray
        Symmetric Mean Absolute Percentage Error.

    References
    ----------
    https://en.wikipedia.org/wiki/Symmetric_mean_absolute_percentage_error

    Examples
    --------
    >>> a = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> b = xr.DataArray(np.random.rand(5, 3, 3), dims=['time', 'x', 'y'])
    >>> xs.smape(a, b, dim='time')
    <xarray.DataArray (x: 3, y: 3)>
    array([[0.35591619, 0.43662087, 0.55372571],
           [0.1864336 , 0.45831965, 0.38473469],
           [0.58730494, 0.18081757, 0.14960832]])
    Dimensions without coordinates: x, y
    """
    dim, axis = _preprocess_dims(dim, a)
    a, b = xr.broadcast(a, b, exclude=dim)
    weights = _preprocess_weights(a, dim, dim, weights)
    input_core_dims = _determine_input_core_dims(dim, weights)

    return xr.apply_ufunc(
        _smape,
        a,
        b,
        weights,
        input_core_dims=input_core_dims,
        kwargs={"axis": axis, "skipna": skipna},
        dask="parallelized",
        output_dtypes=[float],
        keep_attrs=keep_attrs,
    )



import numpy as np
import pytest
from dask import is_dask_collection
from xarray.testing import assert_equal

import xskillscore as xs
from xskillscore import halfwidth_ci_test, mae, sign_test

OFFSET = -1
METRIC = "mae"


@pytest.fixture
def a_1d_worse(a_1d):
    """a_1d worsened by constant offset."""
    return a_1d + OFFSET


@pytest.fixture
def a_1d_worse_less_corr(a_1d):
    """similar but not perfectly correlated a_1d."""
    a_1d_worse = a_1d.copy()
    step = 3
    a_1d_worse[::step] = a_1d[::step].values + 0.1
    return a_1d_worse


@pytest.fixture
def a_worse(a):
    """a worsened by constant offset."""
    return a + OFFSET


def logical(ds):
    """returns results from [0, 1]."""
    return ds > 0.5


@pytest.mark.parametrize("chunk", [True, False])
@pytest.mark.parametrize("input", ["Dataset", "multidim Dataset", "DataArray", "mixed"])
def test_sign_test_inputs(a_1d, a_1d_worse, b_1d, input, chunk):
    """Test sign_test with xr inputs and chunked."""
    if "Dataset" in input:
        name = "var"
        a_1d = a_1d.to_dataset(name=name)
        a_1d_worse = a_1d_worse.to_dataset(name=name)
        b_1d = b_1d.to_dataset(name=name)
        if input == "multidim Dataset":
            a_1d["var2"] = a_1d["var"] * 2
            a_1d_worse["var2"] = a_1d_worse["var"] * 2
            b_1d["var2"] = b_1d["var"] * 2
    elif input == "mixed":
        name = "var"
        a_1d = a_1d.to_dataset(name=name)
    if chunk:
        a_1d = a_1d.chunk()
        a_1d_worse = a_1d_worse.chunk()
        b_1d = b_1d.chunk()
    actual_significantly_different, actual_walk, actual_confidence = sign_test(
        a_1d, a_1d_worse, b_1d, time_dim="time", alpha=0.05, metric="mae"
    )
    # check dask collection preserved
    for actual in [actual_significantly_different, actual_walk, actual_confidence]:
        assert is_dask_collection(actual) if chunk else not is_dask_collection(actual)


@pytest.mark.parametrize("observation", [True, False])
def test_sign_test_categorical(a_1d, a_1d_worse, b_1d, observation):
    """Test sign_test categorical."""
    a_1d = logical(a_1d)
    a_1d_worse = logical(a_1d_worse)
    b_1d = logical(b_1d)
    sign_test(
        a_1d,
        a_1d_worse,
        b_1d,
        time_dim="time",
        metric="categorical",
        orientation="positive",
    )


@pytest.mark.parametrize("metric", ["categorical", "mae"])
def test_sign_test_identical(a_1d, a_1d_worse, b_1d, metric):
    """Test that identical forecasts show no walk step."""
    identicals = [1, 7]
    for i in identicals:
        # set a_1d_worse = a_1d identical for time=i
        a_1d_worse = a_1d_worse.where(a_1d_worse != a_1d_worse.isel(time=i)).fillna(
            a_1d.isel(time=i)
        )
    if metric == "categorical":
        a_1d = logical(a_1d)
        a_1d_worse = logical(a_1d_worse)
        b_1d = logical(b_1d)
    actual_significantly_different, actual_walk, actual_confidence = sign_test(
        a_1d, a_1d_worse, b_1d, time_dim="time", metric=metric, orientation="positive"
    )
    # check flat
    assert (
        actual_walk.diff(dim="time").isel(time=[i - 1 for i in identicals]) == 0
    ).all()


def test_sign_test_alpha(a_1d, a_1d_worse, b_1d):
    """Test that larger alpha leads to small confidence bounds in sign_test."""
    (
        actual_significantly_different_large_alpha,
        actual_walk_large_alpha,
        actual_confidence_large_alpha,
    ) = sign_test(a_1d, a_1d_worse, b_1d, time_dim="time", alpha=0.1, metric="mae")
    (
        actual_significantly_different_small_alpha,
        actual_walk_small_alpha,
        actual_confidence_small_alpha,
    ) = sign_test(a_1d, a_1d_worse, b_1d, time_dim="time", alpha=0.01, metric="mae")
    # check difference in confidence
    assert (actual_confidence_large_alpha < actual_confidence_small_alpha).all()
    # check identical sign_test
    assert actual_walk_large_alpha.equals(actual_walk_small_alpha)


def test_sign_test_user_function(a_1d, a_1d_worse, b_1d):
    """Test sign_test with user function as metric."""

    def mse(a, b, dim):
        return ((a - b) ** 2).mean(dim)

    actual_significantly_different, actual_walk, actual_confidence = sign_test(
        a_1d, a_1d_worse, b_1d, time_dim="time", orientation="negative", metric=mse
    )
    assert actual_walk.isel(time=-1) > 0


@pytest.mark.parametrize("orientation", ["negative", "positive"])
def test_sign_test_orientation(a_1d, a_1d_worse, b_1d, orientation):
    """Test sign_test orientation."""
    actual_significantly_different, actual_walk, actual_confidence = sign_test(
        a_1d, a_1d_worse, b_1d, time_dim="time", orientation=orientation, metric="mae"
    )
    if orientation == "negative":
        # a_1d wins
        assert actual_walk.isel(time=-1) > 0
    elif orientation == "positive":
        # a_1d_worse wins because of corrupted metric orientation
        print(actual_walk.isel(time=-1))
        assert actual_walk.isel(time=-1) < 0


@pytest.mark.parametrize("metric", ["mae", "rmse", "mse"])
def test_sign_test_already_compared_orientation_negative(
    a_1d, a_1d_worse, b_1d, metric
):
    """Test sign_test with forecasts already previously evaluated with observation for
    negative orientation (smaller distances mean better forecast)."""
    a_b_diff = getattr(xs, metric)(a_1d, b_1d, dim=[])
    a_worse_b_diff = getattr(xs, metric)(a_1d_worse, b_1d, dim=[])
    actual_significantly_different, actual_walk, actual_confidence = sign_test(
        a_b_diff, a_worse_b_diff, None, time_dim="time", orientation="negative"
    )
    assert actual_walk.isel(time=-1) > 0


def crpss(o, f_prob, dim=None):
    return 1 - xs.crps_ensemble(o, f_prob, dim=dim) / xs.crps_gaussian(
        o, o.mean("time"), o.std("time"), dim=[]
    )


@pytest.mark.parametrize("metric", [crpss], ids=["crpss"])
def test_sign_test_already_compared_orientation_positive_probabilistic(
    f_prob, o, metric
):
    """Test sign_test for probabilistic crpss metric with positive orientation."""
    o = o.isel(lon=0, lat=0, drop=True)
    f_prob = f_prob.isel(lon=0, lat=0, drop=True)
    f_prob_worse = f_prob + OFFSET
    f_o_diff = metric(o, f_prob, dim=[])
    f_worse_o_diff = metric(o, f_prob_worse, dim=[])
    actual_significantly_different, actual_walk, actual_confidence = sign_test(
        f_o_diff, f_worse_o_diff, None, time_dim="time", orientation="positive"
    )
    assert actual_walk.isel(time=-1) > 0


@pytest.mark.parametrize("metric", ["no_valid_metric_string", (), 1])
def test_sign_test_invalid_metric_fails(metric, a_1d, a_1d_worse, b_1d):
    """Sign_test fails because of invalid metric keyword."""
    with pytest.raises(ValueError) as e:
        sign_test(a_1d, a_1d_worse, b_1d, metric=metric)
    assert "metric" in str(e.value)


def test_sign_test_observations_None_metric_fails(a_1d, a_1d_worse):
    """Sign_test fails because observations None but metric provided."""
    with pytest.raises(ValueError) as e:
        sign_test(a_1d, a_1d_worse, None, metric="mae")
    assert "observations must be provided when metric" in str(e.value)


@pytest.mark.parametrize("orientation", ["categorical", None])
def test_sign_test_invalid_orientation_fails(orientation, a_1d, a_1d_worse, b_1d):
    """Sign_test fails because of invalid orientation keyword."""
    with pytest.raises(ValueError) as e:
        sign_test(a_1d, a_1d_worse, b_1d, orientation=orientation, metric=None)
    assert '`orientation` requires to be either "positive" or' in str(e.value)


@pytest.mark.filterwarnings("ignore:Ignoring provided observation")
def test_sign_test_no_metric_but_observation_warns(a_1d, a_1d_worse, b_1d):
    """Sign_test warns if no metric but observation, ignores observation."""
    (
        actual_significantly_different_with_obs,
        actual_walk_with_obs,
        actual_confidence_with_obs,
    ) = sign_test(a_1d, a_1d_worse, b_1d, orientation="positive", metric=None)
    (
        actual_significantly_different_without_obs,
        actual_walk_without_obs,
        actual_confidence_without_obs,
    ) = sign_test(a_1d, a_1d_worse, None, orientation="positive", metric=None)
    assert (actual_walk_without_obs == actual_walk_with_obs).all()


def test_sign_test_dim(a, a_worse, b):
    """Sign_test with dim specified."""
    actual_significantly_different, actual_walk, actual_confidence = sign_test(
        a,
        a_worse,
        b,
        orientation="positive",
        metric="mse",
        dim=["lon", "lat"],
        time_dim="time",
    )
    # check result reduced by dim
    assert len(actual_walk.dims) == 1


def test_sign_test_dim_fails(a_1d, a_1d_worse, b_1d):
    """Sign_test fails if no time_dim in dim."""
    with pytest.raises(ValueError) as e:
        sign_test(a_1d, a_1d_worse, b_1d, time_dim="time", dim="time")
    assert "`dim` cannot contain `time_dim`" in str(e.value)


def test_sign_test_metric_correlation(a, a_worse, b):
    """Sign_test work for correlation metrics over other dimensions that time_dim."""
    sign_test(a, a_worse, b, time_dim="time", dim=["lon", "lat"], metric="pearson_r")


def test_sign_test_NaNs_confidence(a, a_worse, b):
    """Sign_test confidence with NaNs."""
    actual_significantly_different, actual_walk, actual_confidence = sign_test(
        a, a_worse, b, time_dim="time", metric="mse"
    )
    a_nan = a.copy()
    a_nan[1:3, 1:3, 1:3] = np.nan
    (
        actual_significantly_different_nan,
        actual_walk_nan,
        actual_confidence_nan,
    ) = sign_test(a_nan, a_worse, b, time_dim="time", metric="mse")
    assert not (actual_confidence_nan == actual_confidence).all()


@pytest.mark.parametrize("alpha", [0.05])
@pytest.mark.parametrize("chunk", [True, False])
@pytest.mark.parametrize("input", ["Dataset", "multidim Dataset", "DataArray", "mixed"])
def test_halfwidth_ci_test_inputs(
    a_1d, a_1d_worse_less_corr, b_1d, input, chunk, alpha
):
    """Test halfwidth_ci_test with xr inputs and chunked."""
    if "Dataset" in input:
        name = "var"
        a_1d = a_1d.to_dataset(name=name)
        a_1d_worse_less_corr = a_1d_worse_less_corr.to_dataset(name=name)
        b_1d = b_1d.to_dataset(name=name)
        if input == "multidim Dataset":
            a_1d["var2"] = a_1d["var"] * 2
            a_1d_worse_less_corr["var2"] = a_1d_worse_less_corr["var"] * 2
            b_1d["var2"] = b_1d["var"] * 2
    elif input == "mixed":
        name = "var"
        a_1d = a_1d.to_dataset(name=name)
    if chunk:
        a_1d = a_1d.chunk()
        a_1d_worse_less_corr = a_1d_worse_less_corr.chunk()
        b_1d = b_1d.chunk()
    actual_significantly_different, actual_diff, actual_hwci = halfwidth_ci_test(
        a_1d, a_1d_worse_less_corr, b_1d, METRIC, alpha=alpha
    )
    # check dask collection preserved
    for actual in [actual_significantly_different, actual_diff, actual_hwci]:
        assert is_dask_collection(actual) if chunk else not is_dask_collection(actual)


@pytest.mark.parametrize("alpha", [0.0, 0, 1.0, 1.0, 5.0, 5])
def test_halfwidth_ci_test_alpha(a_1d, a_1d_worse_less_corr, b_1d, alpha):
    """Test halfwidth_ci_test alpha error messages."""
    with pytest.raises(ValueError) as e:
        halfwidth_ci_test(a_1d, a_1d_worse_less_corr, b_1d, METRIC, alpha=alpha)
    assert "`alpha` must be between 0 and 1 or `return_p`" in str(e.value)


@pytest.mark.parametrize("metric", ["mape", "pearson_r"])
def test_halfwidth_ci_test_metric_error(a_1d, a_1d_worse_less_corr, b_1d, metric):
    """Test halfwidth_ci_test alpha error messages."""
    with pytest.raises(ValueError) as e:
        halfwidth_ci_test(a_1d, a_1d_worse_less_corr, b_1d, metric)
    assert "`metric` should be a valid distance metric function" in str(e.value)


@pytest.mark.parametrize("alpha", [0.05])
@pytest.mark.parametrize(
    "metric", ["me", "rmse", "mse", "mae", "median_absolute_error", "smape"]
)
def test_halfwidth_ci_test(a_1d, a_1d_worse_less_corr, b_1d, metric, alpha):
    """Test halfwidth_ci_test favors better forecast."""
    a_1d_worse_less_corr = a_1d.copy()
    # make a_worse worse every second timestep
    step = 3
    a_1d_worse_less_corr[::step] = a_1d_worse_less_corr[::step] + OFFSET * 3
    actual_significantly_different, actual_diff, actual_hwci = halfwidth_ci_test(
        a_1d, a_1d_worse_less_corr, b_1d, metric, alpha=alpha
    )
    assert actual_significantly_different


def test_halfwidth_ci_test_climpred(a_1d, b_1d):
    """Test halfwidth_ci_test as climpred would use it with observations=None."""
    a_1d_worse_less_corr = a_1d.copy()
    # make a_worse worse every second timestep
    a_1d_worse_less_corr[::2] = a_1d_worse_less_corr[::2] + OFFSET
    # calc skill before as in climpred
    dim = []
    time_dim = "time"
    mae_f1o = mae(a_1d, b_1d, dim=dim)
    mae_f2o = mae(a_1d_worse_less_corr, b_1d, dim=dim)

    actual_significantly_different, actual_diff, actual_hwci = halfwidth_ci_test(
        mae_f1o,
        mae_f2o,
        observations=None,
        metric=None,
        dim=dim,
        time_dim=time_dim,
    )
    expected_significantly_different, expected_diff, expected_hwci = halfwidth_ci_test(
        a_1d, a_1d_worse_less_corr, b_1d, METRIC, dim=dim, time_dim=time_dim
    )
    assert_equal(actual_significantly_different, expected_significantly_different)
    assert_equal(actual_diff, expected_diff)
    assert_equal(actual_hwci, expected_hwci)


@pytest.mark.parametrize("dim", [[], ["lon", "lat"]])
def test_halfwidth_ci_test_dim(a, b, dim):
    """Test halfwidth_ci_test for different dim on ."""
    a_worse = a.copy()
    # make a_worse worse every second timestep
    a_worse[::2, :, :] = a_worse[::2, :, :] + OFFSET
    actual_significantly_different, actual_diff, actual_hwci = halfwidth_ci_test(
        a, a_worse, b, METRIC, dim=dim
    )
    # difference larger than half width ci
    assert (actual_diff > actual_hwci).mean() > 0.5
    for d in dim:
        assert d not in actual_diff.dims, print(d, "found, but shouldnt")


def test_halfwidth_ci_test_alpha_hwci(a_1d, a_1d_worse_less_corr, b_1d):
    """Test halfwidth_ci_test with larger alpha leads to smaller hwci."""
    (
        actual_large_alpha_significantly_different,
        actual_large_alpha_diff,
        actual_large_alpha_alpha,
    ) = halfwidth_ci_test(a_1d, a_1d_worse_less_corr, b_1d, METRIC, alpha=0.1)
    (
        actual_small_alpha_significantly_different,
        actual_small_alpha_diff,
        actual_small_alpha_alpha,
    ) = halfwidth_ci_test(a_1d, a_1d_worse_less_corr, b_1d, METRIC, alpha=0.01)
    assert actual_large_alpha_alpha < actual_small_alpha_alpha



# Write the benchmarking functions here.
# See "Writing benchmarks" in the asv docs for more information.

import os

import numpy as np
import xarray as xr

from xskillscore import mae, mse, pearson_r, pearson_r_p_value, rmse

from . import Generate, parameterized, requires_dask

DETERMINISTIC_METRICS = [rmse, pearson_r, mae, mse, pearson_r_p_value]

large_lon_lat = 2000
large_lon_lat_chunksize = large_lon_lat // 4
nmember = 4


class Compute_small(Generate):
    """
    A benchmark xskillscore.metric for small xr.DataArrays"""

    def setup(self, *args, **kwargs):
        self.make_ds(nmember, 1, 1)  # no grid

    @parameterized("metric", DETERMINISTIC_METRICS)
    def time_xskillscore_metric(self, metric):
        """Take time for xskillscore.metric."""
        dim = "member"
        metric(self.ds["tos"], self.ds["sos"], dim=dim).compute()

    @parameterized("metric", DETERMINISTIC_METRICS)
    def peakmem_xskillscore_metric(self, metric):
        dim = "member"
        """Take memory peak for xskillscore.metric."""
        metric(self.ds["tos"], self.ds["sos"], dim=dim).compute()


class Compute_large(Compute_small):
    """
    A benchmark xskillscore.metric for large xr.DataArrays"""

    def setup(self, *args, **kwargs):
        self.make_ds(nmember, large_lon_lat, large_lon_lat)


class Compute_large_dask(Compute_large):
    """
    A benchmark xskillscore.metric for large xr.DataArrays with dask."""

    def setup(self, *args, **kwargs):
        requires_dask()
        self.make_ds(
            nmember,
            large_lon_lat,
            large_lon_lat,
            chunks={"lon": large_lon_lat_chunksize},
        )
